/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TAppDecTop.h
    \brief    Decoder application class (header)
*/

#ifndef __TAPPDECTOP__
#define __TAPPDECTOP__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "../../Lib/TLibVideoIO/TVideoIOYuv.h"
#include "../../Lib/TLibVideoIO/TVideoIOBits.h"
#include "../../Lib/TLibCommon/TComList.h"
#include "../../Lib/TLibCommon/TComPicYuv.h"
#include "../../Lib/TLibCommon/TComBitStream.h"
#include "../../Lib/TLibDecoder/TDecTop.h"
#include "TAppDecCfg.h"

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// decoder application class
class TAppDecTop : public TAppDecCfg
{
private:
  // class interface
  TDecTop                         m_cTDecTop;                     ///< decoder class
  TComBitstream*                  m_apcBitstream;                 ///< bitstream class
#if HHI_NAL_UNIT_SYNTAX
  TVideoIOBitsStartCode           m_cTVideoIOBitstreamFile;       ///< file I/O class
#else
  TVideoIOBits                    m_cTVideoIOBitstreamFile;       ///< file I/O class
#endif
  TVideoIOYuv                     m_cTVideoIOYuvReconFile;        ///< reconstruction YUV class

  // for output control
  Bool                            m_abDecFlag[ MAX_GOP ];         ///< decoded flag in one GOP
  Int                             m_iPOCLastDisplay;              ///< last POC in display order

#if MC_MEMORY_ACCESS_CALC
  Int         m_iNumPredictiveCodedPicture;  /* # of predictively coded pictures */
  Int64       m_i64VMEM_Kbyte    [5]; /* 0: Uncompressed, 1: 32-bit aligned, 2: 64-bit aligned, 3: 128-bit aligned, 4: 256-bit aligned */
  Int64       m_i64VMEM_1X8_Kbyte[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_4X4_Kbyte[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_8X8_Kbyte[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_NXM_Kbyte[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
#if MC_MEMORY_ACCESS_CACHE
  Int64       m_i64VMEM_Kbyte_Cache    [5]; /* 0: Uncompressed, 1: 32-bit aligned, 2: 64-bit aligned, 3: 128-bit aligned, 4: 256-bit aligned */
  Int64       m_i64VMEM_1X8_Kbyte_Cache[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_4X4_Kbyte_Cache[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_8X8_Kbyte_Cache[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
  Int64       m_i64VMEM_NXM_Kbyte_Cache[5]; /* 0: Uncompressed, 1: Compressed 32-bit aligned, 2: Compressed 64-bit aligned, 3: Compressed 128-bit aligned, 4: Compressed 256-bit aligned */
#endif  
#endif // MC_MEMORY_ACCESS_CALC

  // temporary buffer for IBDI
  TComPicYuv                      m_cTempPicYuv;                  ///< temporary buffer for IBDI

public:
  TAppDecTop();
  virtual ~TAppDecTop() {}

  Void  create            (); ///< create internal members
  Void  destroy           (); ///< destroy internal members
  Void  decode            (); ///< main decoding function

#if MC_MEMORY_ACCESS_CALC
  Void      incrementNumPredictiveCodedPicture          ()            { m_iNumPredictiveCodedPicture++;    }
  Int       getNumPredictiveCodedPicture                ()            { return m_iNumPredictiveCodedPicture;  } 
  Void      updateVMEM_Kbyte          ( Int i, Int64 iSize)           { m_i64VMEM_Kbyte[i] += iSize;       }
  Int64     getVMEM_Kbyte             ( Int i)                        { return  m_i64VMEM_Kbyte[i];        }
  Void      updateVMEM_1X8_Kbyte      ( Int i, Int64 iSize)           { m_i64VMEM_1X8_Kbyte[i] += iSize;   }
  Int64     getVMEM_1X8_Kbyte         ( Int i)                        { return  m_i64VMEM_1X8_Kbyte[i];    }
  Void      updateVMEM_4X4_Kbyte      ( Int i, Int64 iSize)           { m_i64VMEM_4X4_Kbyte[i] += iSize;   }
  Int64     getVMEM_4X4_Kbyte         ( Int i)                        { return  m_i64VMEM_4X4_Kbyte[i];    }
  Void      updateVMEM_8X8_Kbyte      ( Int i, Int64 iSize)           { m_i64VMEM_8X8_Kbyte[i] += iSize;   }
  Int64     getVMEM_8X8_Kbyte         ( Int i)                        { return  m_i64VMEM_8X8_Kbyte[i];    }
  Void      updateVMEM_NXM_Kbyte      ( Int i, Int64 iSize)           { m_i64VMEM_NXM_Kbyte[i] += iSize;   }
  Int64     getVMEM_NXM_Kbyte         ( Int i)                        { return  m_i64VMEM_NXM_Kbyte[i];    }
#if MC_MEMORY_ACCESS_CACHE
  Void      updateVMEM_Kbyte_Cache          ( Int i, Int64 iSize)           { m_i64VMEM_Kbyte_Cache[i] += iSize;       }
  Int64     getVMEM_Kbyte_Cache             ( Int i)                        { return  m_i64VMEM_Kbyte_Cache[i];        }
  Void      updateVMEM_1X8_Kbyte_Cache      ( Int i, Int64 iSize)           { m_i64VMEM_1X8_Kbyte_Cache[i] += iSize;   }
  Int64     getVMEM_1X8_Kbyte_Cache         ( Int i)                        { return  m_i64VMEM_1X8_Kbyte_Cache[i];    }
  Void      updateVMEM_4X4_Kbyte_Cache      ( Int i, Int64 iSize)           { m_i64VMEM_4X4_Kbyte_Cache[i] += iSize;   }
  Int64     getVMEM_4X4_Kbyte_Cache         ( Int i)                        { return  m_i64VMEM_4X4_Kbyte_Cache[i];    }
  Void      updateVMEM_8X8_Kbyte_Cache      ( Int i, Int64 iSize)           { m_i64VMEM_8X8_Kbyte_Cache[i] += iSize;   }
  Int64     getVMEM_8X8_Kbyte_Cache         ( Int i)                        { return  m_i64VMEM_8X8_Kbyte_Cache[i];    }
  Void      updateVMEM_NXM_Kbyte_Cache      ( Int i, Int64 iSize)           { m_i64VMEM_NXM_Kbyte_Cache[i] += iSize;   }
  Int64     getVMEM_NXM_Kbyte_Cache         ( Int i)                        { return  m_i64VMEM_NXM_Kbyte_Cache[i];    }
#endif  
#endif // MC_MEMORY_ACCESS_CALC

protected:
  Void  xCreateDecLib     (); ///< create internal classes
  Void  xDestroyDecLib    (); ///< destroy internal classes
  Void  xInitDecLib       (); ///< initialize decoder class

  Void  xWriteOutput      ( TComList<TComPic*>* pcListPic, Bool& rbAlloc ); ///< write YUV to file
  Void  xDeScalePic       ( TComPic* pcPic, TComPicYuv* pcPicD );           ///< descaling of picture
};

#endif

