/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TDecTop.h
    \brief    decoder class (header)
*/

#ifndef __TDECTOP__
#define __TDECTOP__

#include "../TLibCommon/CommonDef.h"
#include "../TLibCommon/TComList.h"
#include "../TLibCommon/TComPicYuv.h"
#include "../TLibCommon/TComPic.h"
#include "../TLibCommon/TComTrQuant.h"

#include "TDecGop.h"
#include "TDecEntropy.h"
#include "TDecSbac.h"
#include "TDecCAVLC.h"

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// decoder class
class TDecTop
{
private:
  Int                     m_iGopSize;
  Bool                    m_bGopSizeSet;
  int                     m_iMaxRefPicNum;
  UInt                    m_uiBalancedCPUs;

  UInt                    m_uiValidPS;
  TComList<TComPic*>      m_cListPic;         //  Dynamic buffer
  TComSPS                 m_cSPS;
  TComPPS                 m_cPPS;
  TComSlice*              m_apcSlicePilot;

  // functional classes
  TComPrediction          m_cPrediction;
  TComTrQuant             m_cTrQuant;
  TDecGop                 m_cGopDecoder;
  TDecSlice               m_cSliceDecoder;
  TDecCu                  m_cCuDecoder;
  TDecEntropy             m_cEntropyDecoder;
  TDecCavlc               m_cCavlcDecoder;
  TDecSbac                m_cSbacDecoder;
  TDecBinCABAC            m_cBinCABAC;
  TDecBinMultiCABAC       m_cBinMultiCABAC;
  TDecBinPIPE             m_cBinPIPE;
  TDecBinMultiPIPE        m_cBinMultiPIPE;
  TDecV2V                 m_cBinV2VwLB;
  TComLoopFilter          m_cLoopFilter;
  TComAdaptiveLoopFilter  m_cAdaptiveLoopFilter;

public:
  TDecTop();
  virtual ~TDecTop();

  Void  create  ();
  Void  destroy ();

  Void  init();
#if MC_MEMORY_ACCESS_CALC
#if MC_MEMORY_ACCESS_CACHE
  Void  decode ( Bool bEos, TComBitstream* pcBitstream, UInt& ruiPOC, TComList<TComPic*>*& rpcListPic, 
	             Double dMemCompressionRatio, Int iMemCompressionUnitWidth,  Int iMemCompressionUnitHeight, 
				 Double dChromaMemCompressionRatio, Int iChromaMemCompressionUnitWidth,  Int iChromaMemCompressionUnitHeight, 
				 Bool *bDecodePBSliceFlag, Int64 *i64CurrentVMEM, Int64 *i64CurrentVMEM_1X8,Int64 *i64CurrentVMEM_4X4, Int64 *i64CurrentVMEM_8X8, Int64 *i64CurrentVMEM_NXM,
				 Int64 *i64CurrentVMEM_Cache, Int64 *i64CurrentVMEM_1X8_Cache, Int64 *i64CurrentVMEM_4X4_Cache, Int64 *i64CurrentVMEM_8X8_Cache, Int64 *i64CurrentVMEM_NXM_Cache);
#else
  Void  decode ( Bool bEos, TComBitstream* pcBitstream, UInt& ruiPOC, TComList<TComPic*>*& rpcListPic, 
	             Double dMemCompressionRatio, Int iMemCompressionUnitWidth,  Int iMemCompressionUnitHeight, 
				 Double dChromaMemCompressionRatio, Int iChromaMemCompressionUnitWidth,  Int iChromaMemCompressionUnitHeight, 
				 Bool *bDecodePBSliceFlag, Int64 *i64CurrentVMEM, Int64 *i64CurrentVMEM_1X8,Int64 *i64CurrentVMEM_4X4, Int64 *i64CurrentVMEM_8X8, Int64 *i64CurrentVMEM_NXM);
#endif				 
#else // MC_MEMORY_ACCESS_CALC
  Void  decode ( Bool bEos, TComBitstream* pcBitstream, UInt& ruiPOC, TComList<TComPic*>*& rpcListPic );
#endif

  Void  deletePicBuffer();

  Void  setBalancedCPUs( UInt ui ) { m_uiBalancedCPUs = ui; }
  UInt  getBalancedCPUs() { return m_cSPS.getBalancedCPUs(); }

protected:
  Void  xGetNewPicBuffer  (TComSlice* pcSlice, TComPic*& rpcPic);
  Void  xUpdateGopSize    (TComSlice* pcSlice);

};// END CLASS DEFINITION TDecTop


#endif // __TDECTOP__

